"""The :mod:`unimatrix.runtime` package provides a common interface to set
up the application runtime environment.
"""
import asyncio
import importlib
import inspect
import pkg_resources
import os
import warnings

import ioc
import ioc.loader

from unimatrix.lib import meta
from . import logging


IS_CONFIGURED = False


def get_entrypoints(name):
    entrypoints = [
        (entry_point.name, entry_point.load())
        for entry_point
        in pkg_resources.iter_entry_points(name)
    ]
    return sorted(entrypoints, key=lambda x: getattr(x[1], 'WEIGHT', 0))


def execute(func, *args, **kwargs):
    """Setup the Unimatrix environment and run `func`."""
    async def f():
        await on('boot')
        try:
            return await func(*args, **kwargs)
        finally:
            await on('shutdown')
    return asyncio.run(f())


@meta.allow_sync
async def setup():
    """Setup all :mod:`unimatrix` components."""
    global IS_CONFIGURED
    if IS_CONFIGURED:
        return

    ioc.setup()
    logging.configure()
    for _, module in get_entrypoints('unimatrix.runtime'):
        on_setup = getattr(module, 'on_setup', None)
        if on_setup is None:
            continue
        await on_setup()\
            if inspect.iscoroutinefunction(on_setup)\
            else on_setup()

    IS_CONFIGURED = True


@meta.allow_sync
async def teardown():
    """Teardown all :mod:`unimatrix` components."""
    for _, module in get_entrypoints('unimatrix.runtime'):
        on_teardown = getattr(module, 'on_teardown', None)
        if on_teardown is None:
            continue
        await on_teardown()\
            if inspect.iscoroutinefunction(on_teardown)\
            else on_teardown()

    ioc.teardown()
    IS_CONFIGURED = False # pylint: disable=unused-variable


@meta.allow_sync
async def on(name, *args, runlevel=None, **kwargs):
    if name == 'boot':
        await setup()
    elif name == 'shutdown':
        await teardown()

    # Before running the event handlers in UNIMATRIX_BOOT_MODULE, run the
    # entrypoints of installed modules.
    for _, module in get_entrypoints(f'unimatrix.runtime.{name}'):
        handler = getattr(module, name, None)
        if handler is None:
            continue
        await handler()\
            if inspect.iscoroutinefunction(handler)\
            else handler()

    qualname = os.getenv('UNIMATRIX_BOOT_MODULE')
    if qualname is None:
        return
    try:
        obj = importlib.import_module(qualname)
    except ImportError:
        warnings.warn(
            f"UNIMATRIX_BOOT_MODULE points to a non-existant symbol: "
            f"{qualname}.")
        return

    if not callable(obj) and not inspect.ismodule(obj):
        raise TypeError(
            "UNIMATRIX_BOOT_MODULE must point to a callable or module."
        )
    if callable(obj):
        return await obj(name, *args, **kwargs)\
            if inspect.iscoroutinefunction(obj)\
            else obj(name, *args, **kwargs)

    f = getattr(obj, name, None)
    if f is None:
        return
    return await f(*args, **kwargs)\
        if inspect.iscoroutinefunction(f)\
        else f(*args, **kwargs)



"""Parse configuration from environment variables."""
import os

from unimatrix.lib import environ


#: Indicates that the application is running in debug mode.
DEBUG = os.getenv('DEBUG') == '1'

#: Specifies which hosts may make requests to application, based on the
#: content of the ``Host`` header.
HTTP_ALLOWED_HOSTS = environ.parselist(
    os.environ, 'HTTP_ALLOWED_HOSTS', sep=';')

#: Indicate if Cross-Origin Resource Sharing (CORS) is enabled for the
#: application.
HTTP_CORS_ENABLED = environ.parsebool(os.environ, 'HTTP_CORS_ENABLED')

#: The allowed origins as determined based on the ``Origin`` header setting
#: for Cross-Origin Resource Sharing (CORS).
HTTP_CORS_ALLOW_ORIGINS = environ.parselist(
    os.environ, 'HTTP_CORS_ALLOW_ORIGINS', sep=';')

#: The allowed request methods for cross-origin requests.
HTTP_CORS_ALLOW_METHODS = environ.parselist(
    os.environ, 'HTTP_CORS_ALLOW_METHODS', sep=':')

#: The allowed request headers setting for Cross-Origin Resource Sharing (CORS).
HTTP_CORS_ALLOW_HEADERS = environ.parselist(
    os.environ, 'HTTP_CORS_ALLOW_HEADERS', sep=':')

#: Indicates if cross-origin responses may include cookies.
HTTP_CORS_ALLOW_CREDENTIALS = environ.parsebool(
    os.environ, 'HTTP_CORS_ALLOW_CREDENTIALS')

#: The list of response headers that may be included in a cross-origin
#: response, separated by a colon.
HTTP_CORS_EXPOSE_HEADERS = environ.parselist(
    os.environ, 'HTTP_CORS_EXPOSE_HEADERS', sep=':')

#: Specifies the maximum time in seconds that a browser may cache a CORS
#: response. Defaults to ``600``.
HTTP_CORS_TTL = environ.parseint(
    os.environ, 'HTTP_CORS_TTL', default=600)

#: The default secret key.
SECRET_KEY = os.getenv('SECRET_KEY')



"""Declares an interface to load editable text configuration."""
import glob
import os

import jinja2
import yaml

from unimatrix.const import ETCDIR
from unimatrix.const import UNIMATRIX_CONFIG_FILE
from .datastructures import DTO


environment = jinja2.Environment( # nosec
    autoescape=False,
    variable_start_string="${",
    variable_end_string="}"
)


def render(text, **params):
    """Renders the editable text configuration `text`."""
    t = environment.from_string(text)
    return t.render(env=os.environ, **params)


def load(path: str, *args, **kwargs) -> dict:
    """Loads the configuration from the specified path."""
    if not os.path.isabs(path):
        path = os.path.join(ETCDIR, path)
    if not os.path.exists(path):
        return None
    files = list(glob.glob(path if not os.path.isdir(path) else f'{path}/*'))
    etc = {}
    for fn in files:
        print(fn, files)
        with open(fn, 'r') as f:
            etc.update(yaml.safe_load(render(f.read(), **kwargs)))
    return DTO.fromdict(etc)


def unimatrix():
    """Loads settings from the Unimatrix Configuration File."""
    cfg = load(UNIMATRIX_CONFIG_FILE)
    include_paths = cfg.get('include')
    exc = ValueError(".include must be a list of strings")
    if include_paths is not None\
    and not isinstance(include_paths, list):
        raise exc

    # Iterate of the items and update the configuration loaded from the
    # default UNIMATRIX_CONFIG_FILE.
    for path in include_paths: # pragma: no cover
        if not isinstance(path, str):
            raise exc
        cfg.update(load(path) or {})

    return cfg


app = unimatrix()



# pylint: skip-file
import os
import unittest

from unimatrix import environ
from unimatrix.conf import settings


class SettingsTestCase(unittest.TestCase):

    def setUp(self):
        settings.destroy()
        self.settings_module = os.environ.pop(
            'UNIMATRIX_SETTINGS_MODULE', None
        )

    def tearDown(self):
        if self.settings_module is not None:
            os.environ['UNIMATRIX_SETTINGS_MODULE'] = self.settings_module

    @settings.with_overrides(FOO=1)
    def test_override_decorator(self):
        self.assertEqual(settings.FOO, 1)

    def test_override_context(self):
        self.assertNotEqual(getattr(settings, 'FOO', None), 1)
        with settings.override(FOO=1):
            self.assertEqual(settings.FOO, 1)

    def test_settings_loads_unimatrix_settings_module(self):
        os.environ['UNIMATRIX_SETTINGS_MODULE'] = 'unimatrix.tests.settings'
        self.assertEqual(getattr(settings, 'BAR', 1), 2)

    def test_settings_has_DEBUG(self):
        self.assertIn('DEBUG', settings.available)

    def test_can_override_environment_variable(self):
        with settings.override(DEBUG=...):
            self.assertEqual(settings.DEBUG, ...)

    def test_settings_precedes_envrionment(self):
        os.environ['UNIMATRIX_SETTINGS_MODULE'] = 'unimatrix.tests.settings'
        self.assertNotEqual(
            settings.HTTP_CORS_TTL,
            environ.HTTP_CORS_TTL
        )



# pylint: skip-file
import os
import tempfile
import unittest

from .. import etc


class EditableTextConfigurationTestCase(unittest.TestCase):

    def test_load(self):
        cfg = etc.load('app.conf')
        self.assertIn('secrets', cfg)
        self.assertIn('include', cfg)
        self.assertEqual(cfg.include[0], 'conf.d')

    def test_load_nonexisting(self):
        cfg = etc.load('does-not-exist.conf')
        self.assertEqual(cfg, None)

    def test_load_fragments(self):
        cfg = etc.load('conf.d')
        self.assertIn('secrets', cfg)
        self.assertIn('foo', cfg.secrets)
        self.assertEqual(cfg.secrets.foo, 'bar')

    def test_render_with_variable(self):
        value = etc.render('${foo}', foo='bar')
        self.assertEqual(value, 'bar')

    def test_render_environment_variable(self):
        os.environ['FOO'] = 'bar'
        value = etc.render('${env.FOO}')
        self.assertEqual(value, 'bar')

    def test_load_simple(self):
        with tempfile.NamedTemporaryFile('w+') as f:
            f.write(
                'foo:\n'
                '  bar: 1'
            )
            f.seek(0)
            data = etc.load(f.name)
        self.assertIn('foo', data)
        self.assertIn('bar', data['foo'])
        self.assertEqual(data['foo']['bar'], 1)

    def test_load_simple(self):
        os.environ['FOO'] = 'bar'
        with tempfile.NamedTemporaryFile('w+') as f:
            f.write(
                'foo: ${env.FOO}\n'
            )
            f.seek(0)
            data = etc.load(f.name)
        self.assertIn('foo', data)
        self.assertEqual(data['foo'], 'bar')



"""Specifies the default logging configuration for Unimatrix
applications.
"""
import copy
import logging.config
import os


DEFAULT_LOG_LEVEL = 'ERROR'

SYSLOG_HOST = os.getenv('SYSLOG_HOST') or 'localhost'

SYSLOG_PORT = int(os.getenv('SYSLOG_PORT') or 5140)

LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'verbose': {
            'format': "%(asctime)s [%(process)d] [%(levelname)s] %(message)s",
            'datefmt': "[%Y-%m-%d %H:%M:%S %z]"
        },
        'simple_syslog': {
            'format': "%(message)s\n"
        }
    },
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
            'formatter': 'verbose'
        },
        'null': {
            'class': 'logging.NullHandler',
        },
        'syslog': {
            'level': 'DEBUG',
            'class': 'logging.handlers.SysLogHandler',
            'facility': 'user',
            'formatter': 'simple_syslog',
            'address' : (SYSLOG_HOST, SYSLOG_PORT),
        },
    },
    'loggers': {
        'unimatrix': {
            'handlers': ['console'],
            'level': os.getenv('LOG_LEVEL', DEFAULT_LOG_LEVEL),
        },
        'unimatrix.io': {
            'handlers': ['console'],
            'level': os.getenv('LOG_LEVEL', DEFAULT_LOG_LEVEL),
            'propagate': False
        },
    }
}


def configure(loggers=None):
    """Configure loggers."""
    defaults = copy.deepcopy(LOGGING)
    defaults['loggers'].update(loggers or {})
    logging.config.dictConfig(defaults)



"""The :mod:`unimatrix.const` module declares all constants defined by the
Unimatrix framework.
"""
import os
import tempfile


# Specifies the default database connection. If
# ``DB_CONNECTION`` is ``None``, then it is up
# to the implementation to decide.
DB_CONNECTION = os.getenv('DB_CONNECTION')

#: The deployment environment.
DEPLOYMENT_ENV = os.getenv('DEPLOYMENT_ENV')

#: Directory holding application configuration files.
ETCDIR = os.path.abspath(os.getenv('APP_ETCDIR') or 'etc/')

RUNDIR = os.getenv('APP_RUNDIR', '/opt/app')

LIBDIR = os.getenv('APP_LIBDIR',
    os.path.join(RUNDIR, 'lib'))

PKIDIR = os.getenv('APP_PKIDIR',
    os.path.join(RUNDIR, 'pki'))

SECDIR = os.getenv('APP_SECDIR',
    '/var/run/secrets/unimatrixone.io')

# Instructs all implementation code to make sure that
# TLS is used with network connections, if possible.
TLS_ENFORCE = not os.getenv('DISABLE_TLS_ENFORCE') == '1'

TMPDIR = tempfile.gettempdir()

#: The location of the Unimatrix Configuration File.
UNIMATRIX_CONFIG_FILE = os.getenv('UNIMATRIX_CONFIG_FILE') or 'app.conf'

VARDIR = os.getenv('APP_VARDIR', '/var/lib/app')
